/*
 * rrlogind - Time Warner RoadRunner logon client daemon -
 * Copyright (C) December, 1998 Sergei Viznyuk <sv@phystech.com>
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <time.h>
#include <signal.h>
#include <setjmp.h>
#include <netinet/in.h>
#include <sys/socket.h>
#include <limits.h>
#include "md5.h"
#include "roadrunner.h"
#include "rrlogind.h"

extern int  DebugFlag;
extern int  LrSocket;
extern jmp_buf env;
extern time_t  LeaseTime;
extern time_t  TimeOut;
extern char rrMsg[MAX_RR_MSG_SIZE];
extern int TrustedServerIP[MAX_TRUSTED_SERVERS];
extern char creds[MAX_CREDS_SIZE];
extern rrOptions_t rrOptions;

int peekfd(int s,time_t tv_usec);
void parseRrMsg(int i);
int PacketSequence=0;
/*****************************************************************************/
void *rrBound()
{
  int i,j,slen;
  struct sockaddr_in saddr;
  unsigned k;
  switch ( sigsetjmp(env,0xffff) )
    {
      case 0: break;
      case SIGALRM: return &rrLogon;
      default: return &rrStop;
    }
  alarm(LeaseTime);
  slen=sizeof(saddr);
  k=1000000*TimeOut;
#if 0
  switch ( peekfd(LrSocket,k) )
    {
      case 1: return &rrLogon;
      case -1:
        if ( DebugFlag )
          syslog(LOG_DEBUG,"error condition on listening socket %d: %m\n",
          LrSocket);
          return &rrStop;
      default:
	  goto trnc;
    }
#endif
  do
    {
      i=time(NULL);
      switch ( peekfd(LrSocket,k) )
	{
	  case 1:
	    if ( DebugFlag )
	      syslog(LOG_DEBUG,"keepalive missed, restarting login session\n");
	    return &rrLogon;
	  case -1:
	    if ( DebugFlag )
	      syslog(LOG_DEBUG,"error condition on listening socket %d: %m\n",
		LrSocket);
	    return &rrStop;
	  default:
	}
      k=1000000*(time(NULL)-i+300);
      i=recvfrom(LrSocket,rrMsg,sizeof(rrMsg),0,(struct sockaddr *)&saddr,&slen);
      if ( i )
	{
          if ( i == -1 )
	    {
	      syslog(LOG_ERR,"recvfrom(%d,..): %m\n",LrSocket);
	      return &rrStop;
	    }
#ifdef DEBUG
	  fprintf(stderr,"%d bytes received from socket %d\n",i,LrSocket);
#endif
        }
      else
	return &rrLogon;
      j=0;
      while ( TrustedServerIP[j] )
	if ( TrustedServerIP[j] == *(int *)&saddr.sin_addr )
	  goto vrfd;
        else
	  j++;
      if ( DebugFlag )
	syslog(LOG_DEBUG,"unauthorized connection attempt from %u.%u.%u.%u\n",
	((unsigned char *)&saddr.sin_addr)[0],
	((unsigned char *)&saddr.sin_addr)[1],
	((unsigned char *)&saddr.sin_addr)[2],
	((unsigned char *)&saddr.sin_addr)[3]);
      continue;
vrfd: parseRrMsg(i);
      switch ( ntohs(*(ushort *)&rrMsg[0]) )
	{
	  case RR_MSG_STATUS_REQ:
	    if ( DebugFlag )
	      syslog(LOG_DEBUG,"keepalive received from %u.%u.%u.%u\n",
	      ((unsigned char *)&saddr.sin_addr)[0],
	      ((unsigned char *)&saddr.sin_addr)[1],
	      ((unsigned char *)&saddr.sin_addr)[2],
	      ((unsigned char *)&saddr.sin_addr)[3]);
	      
	    *(ushort *)&rrMsg[0]=htons(RR_MSG_STATUS_RESP); // msg type
	    *(ushort *)&rrMsg[sizeof(ushort)]=htons(9*sizeof(ushort)+2*sizeof(ulong)+16); // msg length
	    *(ulong *)&rrMsg[2*sizeof(ushort)]=0;  // session id

	    *(ushort *)&rrMsg[2*sizeof(ushort)+sizeof(ulong)]=htons(RR_PARAM_STATUSCODE);
	    *(ushort *)&rrMsg[3*sizeof(ushort)+sizeof(ulong)]=htons(3*sizeof(ushort));
	    *(ushort *)&rrMsg[4*sizeof(ushort)+sizeof(ulong)]=0;

	    *(ushort *)&rrMsg[5*sizeof(ushort)+sizeof(ulong)]=htons(RR_PARAM_STATUSAUTH);
	    *(ushort *)&rrMsg[6*sizeof(ushort)+sizeof(ulong)]=htons(2*sizeof(ushort)+16);
	    *(ulong *)&creds[32]=htonl(++PacketSequence);
	    *(ushort *)&creds[36]=htons(RR_MSG_STATUS_RESP);
	    md5_buffer(creds,38,&rrMsg[7*sizeof(ushort)+sizeof(ulong)]);

	    *(ushort *)&rrMsg[7*sizeof(ushort)+sizeof(ulong)+16]=htons(RR_PARAM_SEQNUM);
	    *(ushort *)&rrMsg[8*sizeof(ushort)+sizeof(ulong)+16]=htons(2*sizeof(ushort)+sizeof(ulong));
	    *(ulong *)&rrMsg[9*sizeof(ushort)+sizeof(ulong)+16]=htonl(PacketSequence);

	    i=sendto(LrSocket,rrMsg,9*sizeof(ushort)+2*sizeof(ulong)+16,0,(struct sockaddr *)&saddr,sizeof(saddr));
	    if ( i == -1 )
	      {
		syslog(LOG_ERR,"sendto(%d,..): %m\n",LrSocket);
		continue;
	      }
#ifdef DEBUG
	    else
	      fprintf(stderr,"%d bytes sent on socket %d\n",i,LrSocket);
#endif
	    if ( i != 9*sizeof(ushort)+2*sizeof(ulong)+16 )
	      {
		syslog(LOG_ERR,"unable to sent %d bytes on socket %d\n",
		9*sizeof(ushort)+2*sizeof(ulong)+16,LrSocket);
		continue;
	      }
	    break;
	  case RR_MSG_RESTART_REQ:
	    if ( DebugFlag )
	      syslog(LOG_DEBUG,"restart request received from %u.%u.%u.%u\n",
	      ((unsigned char *)&saddr.sin_addr)[0],
	      ((unsigned char *)&saddr.sin_addr)[1],
	      ((unsigned char *)&saddr.sin_addr)[2],
	      ((unsigned char *)&saddr.sin_addr)[3]);
	    if ( rrOptions.val[RR_PARAM_RESTARTAUTH] )
	      {
		if ( rrOptions.len[RR_PARAM_RESTARTAUTH] != 16 )
		  {
		    syslog(LOG_ERR,"invalid restart authorization parameter in server request\n");
		    continue;
		  }
	      }
	    else
	      {
		syslog(LOG_ERR,"restart authorization parameter missing in server request\n");
		continue;
	      }
	    if ( rrOptions.val[RR_PARAM_BLINDING] )
	      {
		if ( rrOptions.len[RR_PARAM_BLINDING] != sizeof(ulong) )
		  {
		    syslog(LOG_ERR,"invalid blinding parameter in server request\n");
		    continue;
		  }
	      }
	    else
	      {
		syslog(LOG_ERR,"blinding parameter missing in server request\n");
		continue;
	      }
	    *(ulong *)&creds[32]= *(ulong *)rrOptions.val[RR_PARAM_BLINDING];
	    *(ushort *)&creds[36]=htons(RR_MSG_RESTART_REQ);
	    md5_buffer(creds,38,rrMsg);
	    if ( memcmp(rrMsg,(char *)rrOptions.val[RR_PARAM_RESTARTAUTH],16) )
	      {
	        syslog(LOG_ERR,"restart request authentication failed\n");
	        continue;
	      }
	    else
	      if ( DebugFlag )
	        syslog(LOG_DEBUG,"restart request authentication successful\n");
	    free(rrOptions.val[RR_PARAM_RESTARTAUTH]);
	    rrOptions.val[RR_PARAM_RESTARTAUTH]=0;
	    rrOptions.len[RR_PARAM_RESTARTAUTH]=0;
	    rrOptions.num--;
	    return &rrLogon;
	  default:
	    if ( DebugFlag )
	      syslog(LOG_DEBUG,"received unexpected message type %d on socket %d\n",
	      ntohs(*(ushort *)&rrMsg[0]),LrSocket);
	    continue;
	}
    }
  while ( 1 );
  return &rrLogoff;
}
