/*
 * rrlogind - Time Warner RoadRunner logon client daemon -
 * Copyright (C) December, 1998 Sergei Viznyuk <sv@phystech.com>
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include <setjmp.h>
#include "md5.h"
#include "rrlogind.h"
#include "roadrunner.h"

extern char *ProgramName;
extern int  DebugFlag;
extern char *ServerName;
extern int  ServerName_len;
extern int  RrPort;
extern jmp_buf env;
extern time_t  TimeOut;
extern char *RrPasswd;
extern int  RrPasswd_len;
extern char HashedRrPasswd[16];

int         peekfd(int s,time_t tv_usec);
int         parsePasswdFile();
int         createlostSocket();
char        rrMsg[MAX_RR_MSG_SIZE];
rrOptions_t rrOptions;

struct hostent	*hp = NULL;
int RrSocket = 0;
/*****************************************************************************/
int tcpConnectByName(serverName,serverPort)
char *serverName;
unsigned short serverPort;
{
  struct sockaddr_in saddr;
  int s,i;

  if ( (hp=gethostbyname(serverName)) == NULL )
    {
      syslog(LOG_ERR,"gethostbyname(\"%s\"): %m\n",serverName);
      return -1;
    }
  if ( (s=socket(AF_INET,SOCK_STREAM,0)) == -1 )
    {
      syslog(LOG_ERR,"socket(..): %m\n");
      return -1;
    }
  saddr.sin_family = hp->h_addrtype;
  saddr.sin_port=htons(serverPort);
  i= -1;
  while ( hp->h_addr_list[++i] )
     {
       *(int *)&saddr.sin_addr= *(int *)hp->h_addr_list[i];
#ifdef DEBUG
       fprintf(stderr,"connecting to %s (%u.%u.%u.%u)\n",serverName,
	       ((unsigned char *)&saddr.sin_addr)[0],
	       ((unsigned char *)&saddr.sin_addr)[1],
	       ((unsigned char *)&saddr.sin_addr)[2],
	       ((unsigned char *)&saddr.sin_addr)[3]);
#endif
       errno=0;
       if ( connect(s,(struct sockaddr *)&saddr,sizeof(saddr)) == 0 ) break;
     }
  if ( errno )
    {
      syslog(LOG_ERR,"connect(%d,..): %m\n",s);
      close(s);
      return -1;
    }
#ifdef DEBUG
    fprintf(stderr,"tcp/ip connected to %s (%u.%u.%u.%u)\n",serverName,
    ((unsigned char *)&saddr.sin_addr.s_addr)[0],
    ((unsigned char *)&saddr.sin_addr.s_addr)[1],
    ((unsigned char *)&saddr.sin_addr.s_addr)[2],
    ((unsigned char *)&saddr.sin_addr.s_addr)[3]);
#endif
  return s;
}
/*****************************************************************************/
void parseRrMsg(i)
int i;
{
  ushort type,len;
  register char *p = &rrMsg[2*sizeof(ushort)+sizeof(ulong)];
  char *end = rrMsg+i;
#ifdef DEBUG
  fprintf(stderr,"total=%d  msgtype=%d  msglen=%d  p-rrMsg=%d end-rrMsg=%d\n",
  i,ntohs(*(ushort *)rrMsg),ntohs(((ushort *)rrMsg)[1]),p-rrMsg,end-rrMsg);
#endif
  if ( i > sizeof(rrMsg) )
    {
      syslog(LOG_ERR,"message size exceeds allocated buffer\n");
      return;
    }
  while ( p < end )
    {
      type=ntohs(*(ushort *)p);
      len=ntohs(((ushort *)p)[1]);
#ifdef DEBUG
      fprintf(stderr,"type=%d  len=%d\n",type,len-4);
#endif
      if ( p+len>end )
	{
	  syslog(LOG_ERR,"invalid parameter length in message received\n");
	  break;
	}
      if ( type < 25 )
	{
          if ( rrOptions.len[type] == len-4 )
	    memcpy(rrOptions.val[type],p+4,len-4);
          else
	    {
	      rrOptions.len[type] = len-4;
	      if ( rrOptions.val[type] )
	        free(rrOptions.val[type]);
	      else
	        rrOptions.num++;
	      rrOptions.val[type] = malloc(len-3);
	      memset(rrOptions.val[type],0,len-3);
	      memcpy(rrOptions.val[type],p+4,len-4);
	    }
	}
      else
	syslog(LOG_ERR,"unknown parameter type %d received\n",type);
      p+=len;
    }
#ifdef DEBUG
  fprintf(stderr,"%d options received:\n",rrOptions.num);
  for (i=0;i<25;i++)
    if ( rrOptions.val[i] )
      switch ( i )
        {
	  case RR_PARAM_STATUSCODE:
	  case RR_PARAM_PROTOSEL:
	  case RR_PARAM_REQPORT:
	  case RR_PARAM_HASH:
	  case RR_PARAM_SEQNUM:
	  case RR_PARAM_LOGINPORT:
	  case RR_PARAM_LOGOUTPORT:
	  case RR_PARAM_STATUSPORT:
	  case RR_PARAM_STATUSAUTH:
	  case RR_PARAM_RESTARTAUTH:
	    fprintf(stderr,"i=%-2d  len=%-2d  option = %u\n",
		i,rrOptions.len[i],ntohs(*(ushort *)rrOptions.val[i]));
	    break;
	  case RR_PARAM_SUSPEND:
	    fprintf(stderr,"i=%-2d  len=%-2d  option = %u\n",
		i,rrOptions.len[i],*(unsigned char *)rrOptions.val[i]);
	    break;
	  default:
	    fprintf(stderr,"i=%-2d  len=%-2d  option = \"%s\"\n",
		i,rrOptions.len[i],(char *)rrOptions.val[i]);
	}
#endif
}
/*****************************************************************************/
void releaseRrOptions()
{
  register int i;
  for (i=0;i<25;i++)
    if ( rrOptions.val[i] ) free(rrOptions.val[i]);
  memset(&rrOptions,0,sizeof(rrOptions));
}
/*****************************************************************************/
int doInit()
{
  int i,j=500000;
  releaseRrOptions();
  do
    {
      j+=j;
      if ( RrSocket > 0 ) close(RrSocket);
      i=2;
      while ( (RrSocket=tcpConnectByName(ServerName,RrPort)) == -1 )
        {
	  if ( DebugFlag )
	    syslog(LOG_DEBUG,"unable to connect to %s, will retry in %d secs\n",
	    ServerName,i);
	  sleep(i);
	  i+= i+2;
	}
// Build protocol negotiation message
      *(ushort *)&rrMsg[0]=htons(RR_MSG_PROTONEG_REQ); // msg type
      *(ushort *)&rrMsg[sizeof(ushort)]=htons(12*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN); // msg length
      *(ulong *)&rrMsg[2*sizeof(ushort)]=0;  // session id
  
      *(ushort *)&rrMsg[2*sizeof(ushort)+sizeof(ulong)]=htons(RR_PARAM_CLIENTVER);
      *(ushort *)&rrMsg[3*sizeof(ushort)+sizeof(ulong)]=htons(3*sizeof(ushort));
      *(ushort *)&rrMsg[4*sizeof(ushort)+sizeof(ulong)]=VERSION_ID;
 
      *(ushort *)&rrMsg[5*sizeof(ushort)+sizeof(ulong)]=htons(RR_PARAM_OSID);
      *(ushort *)&rrMsg[6*sizeof(ushort)+sizeof(ulong)]=htons(2*sizeof(ushort)+OS_ID_LEN);
      memcpy(&rrMsg[7*sizeof(ushort)+sizeof(ulong)],OS_ID,OS_ID_LEN);

      *(ushort *)&rrMsg[7*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN]=htons(RR_PARAM_OSVER);
      *(ushort *)&rrMsg[8*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN]=htons(2*sizeof(ushort)+OS_VERSION_LEN);
      memcpy(&rrMsg[9*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN],OS_VERSION,OS_VERSION_LEN);

      *(ushort *)&rrMsg[9*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN]=htons(RR_PARAM_PROTOLIST);
      *(ushort *)&rrMsg[10*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN]=htons(3*sizeof(ushort));
      *(ushort *)&rrMsg[11*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN]=htons(RR_PROTO_CHALLENGE);
 
      i=sendto(RrSocket,rrMsg,12*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN,0,NULL,0);
      if ( i )
	{
	  if ( i == -1 )
	    {
	      syslog(LOG_ERR,"sendto(%d,..): %m\n",RrSocket);
	      continue;
	    }
#ifdef DEBUG
	  fprintf(stderr,"%d bytes sent on socket %d\n",i,RrSocket);
#endif
	}
      if ( i == 12*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN )
	{
	  if ( peekfd(RrSocket,j) ) continue;
	  i=recvfrom(RrSocket,rrMsg,sizeof(rrMsg),0,NULL,0);
	  if ( i )
	    {
	      if ( i == -1 )
		{
		  syslog(LOG_ERR,"recvfrom(%d,..): %m\n",RrSocket);
		  continue;
		}
#ifdef DEBUG
	      fprintf(stderr,"%d bytes received from socket %d\n",i,RrSocket);
#endif
	    }
	  else
	    {
	      if ( DebugFlag )
		syslog(LOG_ERR,"%s dropped connection\n",ServerName);
	      continue;
	    }
	  if ( ntohs(*(ushort *)&rrMsg[0]) == RR_MSG_PROTONEG_RESP )
	    {
	      parseRrMsg(i);
	      if ( rrOptions.val[RR_PARAM_RESPTEXT] )
		syslog(LOG_ERR,"server response text: %s\n",
		(char *)rrOptions.val[RR_PARAM_RESPTEXT]);
	      if ( rrOptions.val[RR_PARAM_STATUSCODE] )
		switch ( ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]) )
		  {
		    case 0:
		      break;
		    case 500:
		      syslog(LOG_ERR,"invalid request or other error\n");
		      continue;
		    default:
		      syslog(LOG_ERR,"uknown status code %d returned\n",
		      ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]));
		      continue;
		  }
	      if ( rrOptions.val[RR_PARAM_PROTOSEL] )
		{
		  if ( ntohs(*(ushort *)rrOptions.val[RR_PARAM_PROTOSEL]) == RR_PROTO_CHALLENGE )
		    {
		      if ( DebugFlag )
			syslog(LOG_DEBUG,"protocol negotiation successfull\n");
		      break;
		    }
		  else
		    syslog(LOG_ERR,"unsupported protocol selection parameter %d\n",
		    ntohs(*(ushort *)rrOptions.val[RR_PARAM_PROTOSEL]));
		}
	      else
		syslog(LOG_ERR,"protocol selection parameter is missing in server response, will retry in %d secs\n",j/1000000);
	    }
	  else
	    syslog(LOG_ERR,"received unexpected message type %d from %s, will retry in %d secs\n",
	    ntohs(*(ushort *)&rrMsg[0]),ServerName,j/1000000);
	}
      else
	syslog(LOG_ERR,"unable to sent %d bytes on socket %d, will retry in %d secs\n",
	12*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN,RrSocket,j/1000000);
    }
  while ( sleep(j/1000000)>=0 );
  close(RrSocket);
  RrSocket=0;
  return 0;
}
/*****************************************************************************/
void *rrInit()
{
  if ( sigsetjmp(env,0xffff) ) return &rrStop;
  alarm(TimeOut);
  if ( parsePasswdFile() || createlostSocket() ) return &rrStop;
  md5_buffer(RrPasswd,RrPasswd_len,HashedRrPasswd);
  if ( doInit() ) return NULL;
  return &rrLogon;
}
