/*
 * rrlogind - Time Warner RoadRunner logon client daemon -
 * Copyright (C) December, 1998 Sergei Viznyuk <sv@phystech.com>
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <setjmp.h>
#include <time.h>
#include "md5.h"
#include "roadrunner.h"
#include "rrlogind.h"

extern int                   DebugFlag,RrSocket;
extern unsigned short        LoginPort,LogoffPort,StatusPort;
extern int                   tcpConnectByName(char *serverName,unsigned short serverPort);
extern char                  creds[MAX_CREDS_SIZE];
extern time_t                TimeOut;
extern rrOptions_t           rrOptions;
extern jmp_buf               env;
extern char                  creds[MAX_CREDS_SIZE];
extern char                  *UserName;
extern int                   UserName_len;
extern char                  *LoginServer;
extern char                  rrMsg[MAX_RR_MSG_SIZE];

int                   peekfd(int s,time_t tv_usec);
void                  parseRrMsg(int i);
/*****************************************************************************/
void *rrLogoff()
{
  int i,j;
  if ( sigsetjmp(env,0xffff) ) return &rrStop;
  alarm(TimeOut);
  j=500000;
  do
    {
      j+=j;
      if ( RrSocket > 0 ) close(RrSocket);
      i=2;
      while ( (RrSocket=tcpConnectByName(LoginServer,LogoffPort)) == -1 )
        {
	  if ( DebugFlag )
	    syslog(LOG_DEBUG,
	    "unable to connect to %s, will retry in %d secs\n",LoginServer,i);
	  sleep(i);
	  i+= i+2;
	}
// Build logout message
      *(ushort *)&rrMsg[0]=htons(RR_MSG_LOGOUT_REQ); // msg type
      *(ushort *)&rrMsg[sizeof(ushort)]=htons(14*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len); // msg length
      *(ulong *)&rrMsg[2*sizeof(ushort)]=0;  // session id
  
      *(ushort *)&rrMsg[2*sizeof(ushort)+sizeof(ulong)]=htons(RR_PARAM_USERNAME);
      *(ushort *)&rrMsg[3*sizeof(ushort)+sizeof(ulong)]=htons(2*sizeof(ushort)+UserName_len);
      memcpy(&rrMsg[4*sizeof(ushort)+sizeof(ulong)],UserName,UserName_len);

      *(ushort *)&rrMsg[4*sizeof(ushort)+sizeof(ulong)+UserName_len]=htons(RR_PARAM_CLIENTVER);
      *(ushort *)&rrMsg[5*sizeof(ushort)+sizeof(ulong)+UserName_len]=htons(3*sizeof(ushort));
      *(ushort *)&rrMsg[6*sizeof(ushort)+sizeof(ulong)+UserName_len]=VERSION_ID;
 
      *(ushort *)&rrMsg[7*sizeof(ushort)+sizeof(ulong)+UserName_len]=htons(RR_PARAM_OSID);
      *(ushort *)&rrMsg[8*sizeof(ushort)+sizeof(ulong)+UserName_len]=htons(2*sizeof(ushort)+OS_ID_LEN);
      memcpy(&rrMsg[9*sizeof(ushort)+sizeof(ulong)+UserName_len],OS_ID,OS_ID_LEN);

      *(ushort *)&rrMsg[9*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+UserName_len]=htons(RR_PARAM_OSVER);
      *(ushort *)&rrMsg[10*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+UserName_len]=htons(2*sizeof(ushort)+OS_VERSION_LEN);
      memcpy(&rrMsg[11*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+UserName_len],OS_VERSION,OS_VERSION_LEN);

      *(ushort *)&rrMsg[11*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len]=htons(RR_PARAM_REASON);
      *(ushort *)&rrMsg[12*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len]=htons(3*sizeof(ushort));
      *(ushort *)&rrMsg[13*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len]=0;
 
      i=sendto(RrSocket,rrMsg,14*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len,0,NULL,0);
      if ( i == -1 )
	{
	  syslog(LOG_ERR,"sendto(%d,..): %m\n",RrSocket);
	  continue;
	}
#ifdef DEBUG
      else
	fprintf(stderr,"%d bytes sent on socket %d\n",i,RrSocket);
#endif
      if ( i == 14*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len )
	{
	  if ( peekfd(RrSocket,j) ) continue;
	  i=recvfrom(RrSocket,rrMsg,sizeof(rrMsg),0,NULL,0);
	  if ( i )
	    {
	      if ( i == -1 )
		{
		  syslog(LOG_ERR,"recvfrom(%d,..): %m\n",RrSocket);
		  continue;
		}
#ifdef DEBUG
	      fprintf(stderr,"%d bytes received from socket %d\n",i,RrSocket);
#endif
	    }
	  else
	    {
	      syslog(LOG_ERR,"rrLogoff: %s dropped connection\n",LoginServer);
	      continue;
	    }
	  switch ( ntohs(*(ushort *)&rrMsg[0]) )
	    {
	      case RR_MSG_LOGOUT_RESP:
	      case RR_MSG_AUTH_RESP:
		parseRrMsg(i);
		if ( rrOptions.val[RR_PARAM_RESPTEXT] )
		  syslog(LOG_ERR,"server response text: %s\n",
		  (char *)rrOptions.val[RR_PARAM_RESPTEXT]);
		if ( rrOptions.val[RR_PARAM_STATUSCODE] )
		  switch ( ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]) )
		    {
		      case 0:
			if ( DebugFlag )
			  syslog(LOG_DEBUG,"logout negotiation successful\n");
			break;
		      case 500:
			syslog(LOG_ERR,"invalid request or other error\n");
			continue;
		      default:
			syslog(LOG_ERR,"uknown status code %d returned\n",
			ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]));
			continue;
		     }
		break;
	      default:
		syslog(LOG_ERR,"received unexpected message type %d from %s, will retry in %d secs\n",
		ntohs(*(ushort *)&rrMsg[0]),LoginServer,j/1000000);
		continue;
	    }
	  break;
	}
      else
	syslog(LOG_ERR,"unable to sent %d bytes on socket %d, will retry in %d secs\n",
	14*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len,RrSocket,j/1000000);
    }
  while ( sleep(j/1000000)>=0 );
  j=500000;
  do
    {
      j+=j;
// Build logout authorization message
      *(ushort *)&rrMsg[0]=htons(RR_MSG_AUTHLOGOUT_REQ); // msg type
      *(ushort *)&rrMsg[sizeof(ushort)]=htons(6*sizeof(ushort)+2*sizeof(ulong)+16); // msg length
      *(ulong *)&rrMsg[2*sizeof(ushort)]=0;  // session id

      *(ushort *)&rrMsg[2*sizeof(ushort)+sizeof(ulong)]=htons(RR_PARAM_CREDS);
      *(ushort *)&rrMsg[3*sizeof(ushort)+sizeof(ulong)]=htons(2*sizeof(ushort)+16);
      i=time(NULL);
      *(ulong *)&creds[32]=htonl(i);
      *(ushort *)&creds[36]=htons(RR_MSG_AUTHLOGOUT_REQ);
      md5_buffer(creds,38,&rrMsg[4*sizeof(ushort)+sizeof(ulong)]);

      *(ushort *)&rrMsg[4*sizeof(ushort)+sizeof(ulong)+16]=htons(RR_PARAM_BLINDING);
      *(ushort *)&rrMsg[5*sizeof(ushort)+sizeof(ulong)+16]=htons(2*sizeof(ushort)+sizeof(ulong));
      *(ulong *)&rrMsg[6*sizeof(ushort)+sizeof(ulong)+16]=htonl(i);

      i=sendto(RrSocket,rrMsg,6*sizeof(ushort)+2*sizeof(ulong)+16,0,NULL,0);
      if ( i == -1 )
	{
	  syslog(LOG_ERR,"sendto(%d,..): %m\n",RrSocket);
	  continue;
	}
#ifdef DEBUG
      else
	fprintf(stderr,"%d bytes sent on socket %d\n",i,RrSocket);
#endif
      if ( i == 6*sizeof(ushort)+2*sizeof(ulong)+16 )
	{
	  if ( peekfd(RrSocket,j) ) continue;
	  i=recvfrom(RrSocket,rrMsg,sizeof(rrMsg),0,NULL,0);
	  if ( i )
	    {
	      if ( i == -1 )
		{
		  syslog(LOG_ERR,"recvfrom(%d,..): %m\n",RrSocket);
		  continue;
		}
#ifdef DEBUG
	      fprintf(stderr,"%d bytes received from socket %d\n",i,RrSocket);
#endif
	    }
	  else
	    {
	      syslog(LOG_ERR,"rrLogoff: %s dropped connection\n",LoginServer);
	      return &rrLogoff;
	    }
	  switch ( ntohs(*(ushort *)&rrMsg[0]) )
	    {
	      case RR_MSG_LOGOUT_RESP:
		parseRrMsg(i);
		if ( rrOptions.val[RR_PARAM_RESPTEXT] )
		  syslog(LOG_ERR,"server response text: %s\n",
		  (char *)rrOptions.val[RR_PARAM_RESPTEXT]);
		if ( rrOptions.val[RR_PARAM_STATUSCODE] )
		  switch ( ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]) )
		    {
		      case 0:
			if ( DebugFlag )
			  syslog(LOG_DEBUG,"logout authorization successful\n");
			break;
		      case 2:
			syslog(LOG_ERR,"incorrect password supplied\n");
			continue;
		      case 500:
			syslog(LOG_ERR,"invalid request or other error\n");
			continue;
		      default:
			syslog(LOG_ERR,"uknown status code %d returned\n",
			ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]));
			continue;
		    }
		break;
	      default:
		syslog(LOG_ERR,"received unexpected message type %d from %s, will retry in %d secs\n",
		ntohs(*(ushort *)&rrMsg[0]),LoginServer,j/1000000);
		continue;
	    }
	  break;
	}
      else
	syslog(LOG_ERR,"unable to sent %d bytes on socket %d, will retry in %d secs\n",
	6*sizeof(ushort)+2*sizeof(ulong)+16,RrSocket,j/1000000);
    }
  while ( sleep(j/1000000)>=0 );
  close(RrSocket);
  RrSocket=0;
  return &rrLogon;
}
