/*
 * rrlogind - Time Warner RoadRunner logon client daemon -
 * Copyright (C) December, 1998 Sergei Viznyuk <sv@phystech.com>
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <setjmp.h>
#include <string.h>
#include <time.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>
#include "md5.h"
#include "rrlogind.h"
#include "roadrunner.h"

extern char           *ProgramName;
extern char           *RrPasswdFile;
extern int            DebugFlag;
extern char           rrMsg[MAX_RR_MSG_SIZE];
extern rrOptions_t    rrOptions;
extern int            RrSocket;
extern char           *UserName;
extern int            UserName_len;
extern char           *ServerName;
extern int            ServerName_len;
extern unsigned short RrPort,LrPort;
extern jmp_buf        env;
extern char           HashedRrPasswd[16];
extern time_t         TimeOut;
extern struct hostent *hp;

char                  *LoginServer,*TrustedServer[MAX_TRUSTED_SERVERS];
int                   TrustedServerIP[MAX_TRUSTED_SERVERS];
unsigned short        LoginPort,LogoffPort,StatusPort;
int                   peekfd(int s,time_t tv_usec);
void                  parseRrMsg(int i);
int                   tcpConnectByName(char *serverName,unsigned short serverPort);
char                  creds[MAX_CREDS_SIZE];
/*****************************************************************************/
int doLogon()
{
  int i,m=0,j=500000;
bgn:
  do
    {
      j+=j;
      if ( RrSocket ) close(RrSocket);
      i=2;
      if ( rrOptions.val[RR_PARAM_LOGINHOST] )
	LoginServer=(char *)rrOptions.val[RR_PARAM_LOGINHOST];
      else
	LoginServer=ServerName;
      if ( rrOptions.val[RR_PARAM_LOGINPORT] )
	LoginPort= ntohs(*(ushort *)rrOptions.val[RR_PARAM_LOGINPORT]);
      else
	LoginPort= RrPort;
      while ( (RrSocket=tcpConnectByName(LoginServer,LoginPort)) == -1 )
        {
	  if ( DebugFlag )
	    syslog(LOG_DEBUG,
	    "unable to connect to %s, will retry in %d secs\n",LoginServer,i);
	  sleep(i);
	  i+= i+2;
	}
// Build login negotiation message
      *(ushort *)&rrMsg[0]=htons(RR_MSG_LOGIN_REQ); // msg type
      *(ushort *)&rrMsg[sizeof(ushort)]=htons(17*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len); // msg length
      *(ulong *)&rrMsg[2*sizeof(ushort)]=0;  // session id
  
      *(ushort *)&rrMsg[2*sizeof(ushort)+sizeof(ulong)]=htons(RR_PARAM_USERNAME);
      *(ushort *)&rrMsg[3*sizeof(ushort)+sizeof(ulong)]=htons(2*sizeof(ushort)+UserName_len);
      memcpy(&rrMsg[4*sizeof(ushort)+sizeof(ulong)],UserName,UserName_len);

      *(ushort *)&rrMsg[4*sizeof(ushort)+sizeof(ulong)+UserName_len]=htons(RR_PARAM_CLIENTVER);
      *(ushort *)&rrMsg[5*sizeof(ushort)+sizeof(ulong)+UserName_len]=htons(3*sizeof(ushort));
      *(ushort *)&rrMsg[6*sizeof(ushort)+sizeof(ulong)+UserName_len]=VERSION_ID;
 
      *(ushort *)&rrMsg[7*sizeof(ushort)+sizeof(ulong)+UserName_len]=htons(RR_PARAM_OSID);
      *(ushort *)&rrMsg[8*sizeof(ushort)+sizeof(ulong)+UserName_len]=htons(2*sizeof(ushort)+OS_ID_LEN);
      memcpy(&rrMsg[9*sizeof(ushort)+sizeof(ulong)+UserName_len],OS_ID,OS_ID_LEN);

      *(ushort *)&rrMsg[9*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+UserName_len]=htons(RR_PARAM_OSVER);
      *(ushort *)&rrMsg[10*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+UserName_len]=htons(2*sizeof(ushort)+OS_VERSION_LEN);
      memcpy(&rrMsg[11*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+UserName_len],OS_VERSION,OS_VERSION_LEN);

      *(ushort *)&rrMsg[11*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len]=htons(RR_PARAM_REASON);
      *(ushort *)&rrMsg[12*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len]=htons(3*sizeof(ushort));
      *(ushort *)&rrMsg[13*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len]=0;
 
      *(ushort *)&rrMsg[14*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len]=htons(RR_PARAM_REQPORT);
      *(ushort *)&rrMsg[15*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len]=htons(3*sizeof(ushort));
      *(ushort *)&rrMsg[16*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len]=htons(LrPort);
 
      i=sendto(RrSocket,rrMsg,17*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len,0,NULL,0);
      if ( i )
	{
	  if ( i == -1 )
	    {
	      syslog(LOG_ERR,"sendto(%d,..): %m\n",RrSocket);
	      continue;
	    }
#ifdef DEBUG
	  fprintf(stderr,"%d bytes sent on socket %d\n",i,RrSocket);
#endif
	}
      if ( i == 17*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len )
	{
	  if ( peekfd(RrSocket,j) ) continue;
	  i=recvfrom(RrSocket,rrMsg,sizeof(rrMsg),0,NULL,0);
	  if ( i )
	    {
	      if ( i == -1 )
		{
		  syslog(LOG_ERR,"recvfrom(%d,..): %m\n",RrSocket);
		  continue;
		}
#ifdef DEBUG
	      fprintf(stderr,"%d bytes received from socket %d\n",i,RrSocket);
#endif
	    }
	  else
	    {
	      syslog(LOG_ERR,"%s dropped connection\n",LoginServer);
	      continue;
	    }
	  switch ( ntohs(*(ushort *)&rrMsg[0]) )
	    {
	      case RR_MSG_AUTH_RESP:
	      case RR_MSG_LOGIN_RESP:
		parseRrMsg(i);
		if ( rrOptions.val[RR_PARAM_RESPTEXT] )
		  syslog(LOG_INFO,"server response text: %s\n",
		  (char *)rrOptions.val[RR_PARAM_RESPTEXT]);
		if ( rrOptions.val[RR_PARAM_STATUSCODE] )
		  switch ( ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]) )
		    {
		      case 0:
			if ( DebugFlag )
			  syslog(LOG_DEBUG,"login negotiation successful\n");
			break;
		      case 500:
			syslog(LOG_ERR,"invalid request or other error\n");
			continue;
		      case 501:
			syslog(LOG_ERR,"logon server is hosed\n");
			continue;
		      default:
			syslog(LOG_ERR,"uknown status code %d returned\n",
			ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]));
			continue;
		     }
		break;
	      default:
		syslog(LOG_ERR,"received unexpected message type %d from %s, will retry in %d secs\n",
		ntohs(*(ushort *)&rrMsg[0]),LoginServer,j/1000000);
		continue;
	    }
	  if ( rrOptions.val[RR_PARAM_HASH] )
	    {
	      if ( ! *(ushort *)rrOptions.val[RR_PARAM_HASH] )
		{
		  syslog(LOG_ERR,"request to send plain text password received from %s\n",LoginServer);
		  return -1;
		}
	    }
	  else
	    {
	      syslog(LOG_ERR,"hash parameter missing in server response, will retry in %d secs\n",j/1000000);
	      continue;
	    }
	  if ( rrOptions.val[RR_PARAM_NONCE] )
	    {
	      if ( rrOptions.len[RR_PARAM_NONCE] == 16 ) break;
	      syslog(LOG_ERR,"invalid nonce parameter in server response, will retry in %d secs\n",j/1000000);
	    }
	  else
	    syslog(LOG_ERR,"nonce parameter missing in server response, will retry in %d secs\n",j/1000000);
	}
      else
	syslog(LOG_ERR,"unable to sent %d bytes on socket %d, will retry in %d secs\n",
	17*sizeof(ushort)+sizeof(ulong)+OS_ID_LEN+OS_VERSION_LEN+UserName_len,RrSocket,j/1000000);
    }
  while ( sleep(j/1000000)>=0 );
  j=1000000;
  do
    {
      j+=j;
// Build login authorization message
      *(ushort *)&rrMsg[0]=htons(RR_MSG_AUTHLOGIN_REQ); // msg type
      *(ushort *)&rrMsg[sizeof(ushort)]=htons(6*sizeof(ushort)+2*sizeof(ulong)+16); // msg length
      *(ulong *)&rrMsg[2*sizeof(ushort)]=0;  // session id

      *(ushort *)&rrMsg[2*sizeof(ushort)+sizeof(ulong)]=htons(RR_PARAM_CREDS);
      *(ushort *)&rrMsg[3*sizeof(ushort)+sizeof(ulong)]=htons(2*sizeof(ushort)+16);
      memset(creds,0,256);
      memcpy(creds,(char *)rrOptions.val[RR_PARAM_NONCE],16);
      memcpy(&creds[16],HashedRrPasswd,16);
      i=time(NULL);
      *(ulong *)&creds[32]=htonl(i);
      *(ushort *)&creds[36]=htons(RR_MSG_AUTHLOGIN_REQ);
      md5_buffer(creds,38,&rrMsg[4*sizeof(ushort)+sizeof(ulong)]);

      *(ushort *)&rrMsg[4*sizeof(ushort)+sizeof(ulong)+16]=htons(RR_PARAM_BLINDING);
      *(ushort *)&rrMsg[5*sizeof(ushort)+sizeof(ulong)+16]=htons(2*sizeof(ushort)+sizeof(ulong));
      *(ulong *)&rrMsg[6*sizeof(ushort)+sizeof(ulong)+16]=htonl(i);

      i=sendto(RrSocket,rrMsg,6*sizeof(ushort)+2*sizeof(ulong)+16,0,NULL,0);
      if ( i )
	{
	  if ( i == -1 )
	    {
	      syslog(LOG_ERR,"rrLogon: sendto(%d,..): %m\n",RrSocket);
	      goto bgn;
	    }
#ifdef DEBUG
	  fprintf(stderr,"%d bytes sent on socket %d\n",i,RrSocket);
#endif
	}
      if ( i == 6*sizeof(ushort)+2*sizeof(ulong)+16 )
	{
	  if ( peekfd(RrSocket,j) ) continue;
	  i=recvfrom(RrSocket,rrMsg,sizeof(rrMsg),0,NULL,0);
	  if ( i )
	    {
	      if ( i == -1 )
		{
		  syslog(LOG_ERR,"rrLogon: recvfrom(%d,..): %m\n",RrSocket);
		  goto bgn;
		}
#ifdef DEBUG
	      fprintf(stderr,"%d bytes received from socket %d\n",i,RrSocket);
#endif
	    }
	  else
	    {
	      syslog(LOG_ERR,"rrLogon: %s dropped connection\n",LoginServer);
	      goto bgn;
	    }
	  switch ( ntohs(*(ushort *)&rrMsg[0]) )
	    {
	      case RR_MSG_LOGIN_RESP:
		parseRrMsg(i);
		if ( rrOptions.val[RR_PARAM_RESPTEXT] )
		  syslog(LOG_ERR,"server response text: %s\n",
		  (char *)rrOptions.val[RR_PARAM_RESPTEXT]);
		if ( rrOptions.val[RR_PARAM_STATUSCODE] )
		  switch ( ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]) )
		    {
		      case 0:
			if ( DebugFlag )
			  syslog(LOG_DEBUG,"login authorization successful\n");
			break;
		      case 2:
			syslog(LOG_ERR,"rrLogon: incorrect password supplied\n");
			return -1;
		      case 500:
			syslog(LOG_ERR,"rrLogon: invalid request or other error\n");
			goto bgn;
		      case 501:
			syslog(LOG_ERR,"rrLogon: logon server is hosed, will retry in %d secs\n",j/1000000);
			continue;
		      default:
			syslog(LOG_ERR,"rrLogon: uknown status code %d returned\n",
			ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSCODE]));
			goto bgn;
		     }
		break;
	      default:
		syslog(LOG_ERR,"rrLogon: received unexpected message type %d from %s\n",ntohs(*(ushort *)&rrMsg[0]),LoginServer);
		goto bgn;
	    }
	  if ( rrOptions.val[RR_PARAM_LOGINHASH] )
	    {
	      if ( rrOptions.len[RR_PARAM_LOGINHASH] != 16 )
		{
		  syslog(LOG_ERR,"rrLogon: invalid login hash parameter in server response\n");
		  goto bgn;
		}
	    }
	  else
	    {
	      syslog(LOG_ERR,"rrLogon: login hash parameter missing in server response\n");
	      goto bgn;
	    }
	  memcpy(&creds[32],&rrMsg[8],i-28);
	  *(ushort *)&creds[i+4]=htons(RR_MSG_LOGIN_RESP);
	  md5_buffer(creds,i+6,rrMsg);
	  if ( memcmp(rrMsg,(char *)rrOptions.val[RR_PARAM_LOGINHASH],16) )
	    {
	      syslog(LOG_ERR,"rrLogon: login authentication failed\n");
	      return -1;
	    }
	  else
	    if ( DebugFlag )
	      syslog(LOG_DEBUG,"login authentication successful\n");
	}
      else
	{
	  syslog(LOG_ERR,"rrLogon: unable to sent %d bytes on socket %d\n",
	  6*sizeof(ushort)+2*sizeof(ulong)+16,RrSocket);
	  goto bgn;
	}
      if ( rrOptions.val[RR_PARAM_LOGOUTPORT] )
	LogoffPort=ntohs(*(ushort *)rrOptions.val[RR_PARAM_LOGOUTPORT]);
      else
	{
	  if ( DebugFlag)
	    syslog(LOG_DEBUG,"logout port parameter missing, assuming %d\n",
	  LoginPort);
	  LogoffPort=LoginPort;
        }
      if ( rrOptions.val[RR_PARAM_STATUSPORT] )
	StatusPort=ntohs(*(ushort *)rrOptions.val[RR_PARAM_STATUSPORT]);
      else
	{
	  if ( DebugFlag)
	    syslog(LOG_DEBUG,"status port parameter missing, assuming %d\n",
	    LoginPort);
	  StatusPort=LoginPort;
	}
      memset(TrustedServer,0,sizeof(TrustedServer));
      memset(TrustedServerIP,0,sizeof(TrustedServerIP));
      if ( rrOptions.val[RR_PARAM_TRUSTEDSERVERS] )
	{
	  char *s=(char *)rrOptions.val[RR_PARAM_TRUSTEDSERVERS];
	  char *e=(char *)rrOptions.val[RR_PARAM_TRUSTEDSERVERS]+rrOptions.len[RR_PARAM_TRUSTEDSERVERS];
	  i=0;
	  do
	    {
	      TrustedServer[i]=s;
	      while ( *s > 44 ) s++;
	      *s++ =0;
	      i++;
	    }
	  while ( (s<e) && (i<MAX_TRUSTED_SERVERS-1) );
	}
      else
	{
	  if ( DebugFlag )
	    syslog(LOG_DEBUG,"trusted servers parameter missing, assuming %s\n",
	    LoginServer);
	  TrustedServer[0]=LoginServer;
	}
      i=0;
      while ( TrustedServer[i] )
	{
	  if ( (hp=gethostbyname(TrustedServer[i])) )
	    {
	      int n=0;
	      while ( hp->h_addr_list[n] )
	      TrustedServerIP[m++]= *(int *)hp->h_addr_list[n++];
	    }
	  else
	    if ( DebugFlag )
	      syslog(LOG_DEBUG,"gethostbyname(\"%s\"): %m\n",TrustedServer[i]);
	  i++;
	}
#ifdef DEBUG
      i= -1;
      while ( TrustedServer[++i] )
	fprintf(stderr,"trusted server %d: %s (%u.%u.%u.%u)\n",
	i+1,TrustedServer[i],
	((unsigned char *)&TrustedServerIP[i])[0],
	((unsigned char *)&TrustedServerIP[i])[1],
	((unsigned char *)&TrustedServerIP[i])[2],
	((unsigned char *)&TrustedServerIP[i])[3]);
#endif
      close(RrSocket);
      RrSocket=0;
      return 0;
    }
  while ( sleep(j/1000000)>=0 );
  close(RrSocket);
  RrSocket=0;
  return 0;
}
/*****************************************************************************/
void *rrLogon()
{
  if ( sigsetjmp(env,0xffff) ) return &rrStop;
  alarm(TimeOut);
  if ( doLogon() ) return &rrInit;
  return &rrBound;
}
