/*
 * rrlogind - Time Warner RoadRunner logon client daemon -
 * Copyright (C) December, 1998 Sergei Viznyuk <sv@phystech.com>
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <syslog.h>
#include <unistd.h>
#include <errno.h>
#include <ctype.h>
#include <string.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include "rrlogind.h"
#include "md5.h"

extern char *ProgramName;
extern char *RrPasswdFile;
extern int  DebugFlag;
extern int  RrSocket;

char        HashedRrPasswd[16];
char        *ServerName         =       NULL;
int	    ServerName_len      =       0;
char	    *UserName           =       NULL;
int	    UserName_len        =       0;
int	    RrPort              =       0;
char        *RrPasswd	        =	NULL;
int         RrPasswd_len	=	0;
unsigned short LrPort           =       LOSTSOCKMINPORT;
int         LrSocket            =       0;
int         doInit();
int         doLogon();
/*****************************************************************************/
int parsePasswdFile()
{
  FILE *fp;
  char *a;
  static char buf[64];
  register char *c;
  char *server = NULL;
  int server_len = 0;

  fp=fopen(RrPasswdFile,"r");
  if ( fp == NULL )
    {
      fprintf(stderr,"****  %s: cannot open \"%s\": %s\n",
	      ProgramName,RrPasswdFile,sys_errlist[errno]);
      return -1;
    }
  while ( fgets(buf,64,fp) )
    {
      c=buf;
      while ( isspace(*c) ) c++;
      if ( *c == '#' ) continue;
      a=c;
      while ( *c )
	if ( *c == ':' )
	  break;
        else
	  c++;
      if ( *c == 0 ) continue;
      *c = 0;
      if ( ServerName )
	{
	  if ( memcmp(a,ServerName,ServerName_len) ) continue;
        }
      else
	{
	  server=a;
	  server_len = c-a;
	}
      a= ++c;
      while ( *c )
	if ( *c == ':' )
	  break;
        else
	  c++;
      if ( *c == 0 ) continue;
      *c = 0;
      if ( RrPort )
	{
          if ( atoi(a) )
	    {
	      if ( RrPort-atoi(a) ) continue;
	    }
	  else
	    {
	      if ( RrPort != DEFAULT_PORT ) continue;
	    }
	}
      else
	if ( atoi(a) )
	  RrPort=atoi(a);
	else
	  RrPort=DEFAULT_PORT;
      a= ++c;
      while ( *c )
	if ( *c == ':' )
	  break;
        else
	  c++;
      if ( *c == 0 ) continue;
      *c = 0;
      if ( UserName )
	{
	  if ( memcmp(a,UserName,UserName_len) ) continue;
        }
      else
	{
	  UserName=a;
	  UserName_len = c-a;
	}
      a= ++c;
      RrPasswd= a;
      while ( *c > 32 ) c++;
      *c = 0;
      RrPasswd_len = c-a;
      if ( ServerName == NULL )
	{
	  ServerName=server;
	  ServerName_len=server_len;
	}
#ifdef DEBUG
      fprintf(stderr,"\
RrPasswdFile = \"%s\"\n\
ServerName = \"%s\"   ServerName_len=%d\n\
RrPort = %d\n\
UserName = \"%s\"   UserName_len=%d\n\
RrPasswd = \"%s\"   RrPasswd_len=%d\n",
RrPasswdFile,ServerName,ServerName_len,RrPort,
UserName,UserName_len,RrPasswd,RrPasswd_len);
#endif
      break;
    }
  fclose(fp);
  if ( ServerName )
    {
      if ( UserName )
        {
	  if ( RrPasswd )
	    return 0;
          else
            fprintf(stderr,"****  %s: no password",ProgramName);
	}
      else
        fprintf(stderr,"****  %s: no username",ProgramName);
    }
  else
    fprintf(stderr,"****  %s: no server",ProgramName);
  fprintf(stderr," found or specified\n");
  return -1;
}
/*****************************************************************************/
int createlostSocket()
{
  struct sockaddr_in addr;

  LrSocket=socket(AF_INET,SOCK_DGRAM,0);
  if ( LrSocket == -1 )
    {
      syslog(LOG_ERR,"socket(..): %m\n");
      return -1;
    }
  memset(&addr,0,sizeof(struct sockaddr_in));
  addr.sin_family = AF_INET;
  addr.sin_addr.s_addr = INADDR_ANY;
  do
    {
      LrPort++;
      addr.sin_port = htons(LrPort);
      errno = 0;
      bind(LrSocket,(struct sockaddr *)&addr,sizeof(struct sockaddr_in));
    }
  while ( errno == EADDRINUSE );
  if ( errno )
    {
      syslog(LOG_ERR,"bind(%d,..): %m\n",LrSocket);
      close(LrSocket);
      return -1;
    }
#if 0
  if ( listen(LrSocket,1) == -1 )
    {
      syslog(LOG_ERR,"listen(%d,1): %m\n",LrSocket);
      close(LrSocket);
      return -1;
    }
#endif
#ifdef DEBUG
  fprintf(stderr,"created listening LrSocket=%d on port %d\n",LrSocket,LrPort);
#endif
   return 0;
}
/*****************************************************************************/
void *rrStart()
{
  if ( parsePasswdFile() || createlostSocket() ) return NULL;
  md5_buffer(RrPasswd,RrPasswd_len,HashedRrPasswd);
  if ( doInit() ) return NULL;
  if ( doLogon() ) return NULL;
  return &rrBound;
}
