/*
 * rrlogind - Road Runner login client daemon -
 * Copyright (C) January, 1998 Sergei Viznyuk <sv@phystech.com>
 * 
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <sys/types.h>
#include <stdio.h>
#include <unistd.h>
#include <syslog.h>
#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#include "rrlogind.h"

extern char		*ProgramName;
extern int		DebugFlag;
extern void		*(*currState)();
jmp_buf		        env;
/*****************************************************************************/
void killPid(sig)
int sig;
{
  FILE *fp;
  pid_t pid;
  fp=fopen(PID_FILE_PATH,"r");
  if ( fp == NULL ) goto ntrn;
  fscanf(fp,"%u",&pid);
  fclose(fp);
  if ( kill(pid,sig) )
    {
ntrn: fprintf(stderr,"****  %s: not running\n",ProgramName);
      exit(1);
    }
  exit(0);
}
/*****************************************************************************/
void writePidFile()
{
  FILE *fp;
  fp=fopen(PID_FILE_PATH,"w");
  if ( fp == NULL )
    {
      syslog(LOG_ERR,"writePidFile: fopen: %m\n");
      exit(1);
    }
  fprintf(fp,"%u\n",getpid());
  fclose (fp);
}
/*****************************************************************************/
void checkIfAlreadyRunning()
{
  int o;
  o=open(PID_FILE_PATH,O_RDONLY);
  if ( o == -1 ) return;
  close(o);
  fprintf(stderr,"\
****  %s: already running\n\
****  %s: if not then delete "PID_FILE_PATH" file\n",ProgramName,ProgramName);
  exit(1);
}
/*****************************************************************************/
void sigHandler(sig)
int sig;
{
  switch ( sig )
    {
     case SIGALRM:
     case SIGHUP:
     case SIGPIPE:
      if ( currState == &rrBound )
        siglongjmp(env,sig); /* lease time out */
      else
	{
	  if ( currState == &rrLogoff )
	    siglongjmp(env,sig); /* timed out trying to logoff */
	  else
	    {
	      if ( currState == &rrLogon )
		siglongjmp(env,sig); /* timed out trying to logon */
	      else
		{
		  if ( currState == &rrInit )
		    siglongjmp(env,sig); /* timed out trying to negitiate protocol */
	          else
		    syslog(LOG_ERR,"timed out trying to login\n");
		}
	    }
	}
       break;
      case SIGSEGV:
        syslog(LOG_ERR,"terminating on signal %d\n",sig);
  	unlink(PID_FILE_PATH);
	exit(sig);	
      default:
       syslog(LOG_ERR,"terminating on signal %d\n",sig);
    }
  rrStop();
  unlink(PID_FILE_PATH);
  exit(sig);
}
/*****************************************************************************/
void signalSetup()
{
  int i;
  struct sigaction action;
  sigaction(SIGHUP,NULL,&action);
  action.sa_handler= &sigHandler;
  action.sa_flags = 0;
  for (i=1;i<16;i++) sigaction(i,&action,NULL);
  action.sa_handler= SIG_IGN;
  sigaction(SIGCHLD,&action,NULL);
}
